package edu.ucla.sspace.graph;

import edu.ucla.sspace.common.Similarity;

import edu.ucla.sspace.clustering.Assignment;
import edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering;
import edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering.ClusterLinkage;
import edu.ucla.sspace.clustering.Merge;
import edu.ucla.sspace.clustering.SoftAssignment;

import edu.ucla.sspace.matrix.AbstractMatrix;
import edu.ucla.sspace.matrix.Matrix;
import edu.ucla.sspace.matrix.SparseHashMatrix;
import edu.ucla.sspace.matrix.SparseMatrix;
import edu.ucla.sspace.matrix.SparseSymmetricMatrix;
import edu.ucla.sspace.matrix.YaleSparseMatrix;

import edu.ucla.sspace.util.Counter;
import edu.ucla.sspace.util.HashMultiMap;
import edu.ucla.sspace.util.HashIndexer;
import edu.ucla.sspace.util.Indexer;
import edu.ucla.sspace.util.MultiMap;
import edu.ucla.sspace.util.ObjectCounter;
import edu.ucla.sspace.util.Pair;
import edu.ucla.sspace.util.WorkQueue;

import edu.ucla.sspace.util.primitive.IntIterator;
import edu.ucla.sspace.util.primitive.IntSet;
import edu.ucla.sspace.util.primitive.IntIntMultiMap;
import edu.ucla.sspace.util.primitive.IntIntHashMultiMap;

import edu.ucla.sspace.vector.CompactSparseIntegerVector;
import edu.ucla.sspace.vector.DoubleVector;
import edu.ucla.sspace.vector.SparseDoubleVector;
import edu.ucla.sspace.vector.SparseHashIntegerVector;
import edu.ucla.sspace.vector.SparseIntegerVector;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Properties;
import java.util.Queue;
import java.util.Set;

import java.util.concurrent.ConcurrentNavigableMap;
import java.util.concurrent.ConcurrentSkipListMap;

import java.util.logging.Level;
import java.util.logging.Logger;

import static edu.ucla.sspace.util.LoggerUtil.verbose;
import static edu.ucla.sspace.util.LoggerUtil.veryVerbose;


/**
 * @author David Jurgens 
 */
public class RelationClustering implements java.io.Serializable {

    public static edu.ucla.sspace.util.Indexer<?> indexer;
    
    private static final long serialVersionUID = 1L;

    /**
     * A prefix for specifying properties.
     */
    private static final String PROPERTY_PREFIX =
        "edu.ucla.sspace.clustering.RelationClustering";

    /**
     * The property to specify if the edge similarity matrix should be kept in
     * memory during clustering, or if its values should be computed on the fly.
     */
    public static final String KEEP_SIMILARITY_MATRIX_IN_MEMORY_PROPERTY =
        PROPERTY_PREFIX + ".keepSimilarityMatrixInMemory";
    
    /**
     * The logger to which clustering status updates will be written.
     */
    private static final Logger LOGGER =
        Logger.getLogger(RelationClustering.class.getName());

    /**
     * The work used by all {@code LinkClustering} instances to perform
     * multi-threaded operations.
     */
    private static final WorkQueue WORK_QUEUE = WorkQueue.getWorkQueue(8);

    /**
     * Instantiates a new {@code RelationClustering} instance.
     */
    public RelationClustering() { }

    /**
     * <i>Ignores the specified number of clusters</i> and returns the
     * clustering solution according to the partition density.
     *
     * @param numClusters this parameter is ignored.
     */
    public <T,E extends TypedEdge<T>> MultiMap<Integer,Integer> cluster(
                        Multigraph<T,E> graph, RelationSemantics<T> semantics, 
                        Properties props) {

        return singleLink(graph, semantics);
    }

    /**
     * Computes the similarity of the two edges as the Tanimoto coefficient of
     * the multisets of the edges connecting the neighbors of two impost nodes.
     * The impost nodes are the two nodes the edges do not have in common.
     */
    protected <T,E extends TypedEdge<T>> double getConnectionSimilarity(
            Multigraph<T,E> graph, int keystone, int impost1, int impost2,
            RelationSemantics<T> simFunc) {
        
        int verts = graph.order() + 1;       
        SparseIntegerVector neighbors1 = new SparseHashIntegerVector(verts);
        SparseIntegerVector neighbors2 = new SparseHashIntegerVector(verts);
        neighbors1.add(impost1, 1);
        neighbors2.add(impost2, 1);

        double edgeSimSum = 0d;

        // Find all the neighbors of the first impost node
        for (E e1 : graph.getAdjacencyList(impost1)) {
            int neighbor = (impost1 == e1.to()) ? e1.from() : e1.to();
                            
            neighbors1.add(neighbor, 1);
            // Examine all the other edges that the other impost node may share
            // as neighbors.  Note that this iteration will be a no-op if the
            // other impost does not have any edges to this neighbor.
            for (E e2 : graph.getEdges(impost2, neighbor)) 
                edgeSimSum += simFunc.similarity(e1, e2);

            // SPECIAL CASE: if the neighbor is to the other node, then assume
            // maximum similarity.  Note that in a full version that takes into
            // account node semantics, this would need to take into account the
            // node similarity
            if (neighbor == impost2)
                edgeSimSum++;
        }

        // Find all the neighbors for the second impost
        for (E e2 : graph.getAdjacencyList(impost2)) {
            int neighbor = (impost2 == e2.to()) ? e2.from() : e2.to();
           
            neighbors2.add(neighbor, 1);
            // SPECIAL CASE: if the neighbor is to the other node, then assume
            // maximum similarity.  Note that in a full version that takes into
            // account node semantics, this would need to take into account the
            // node similarity
            if (neighbor == impost1)
                edgeSimSum++;
        }
        
        // Compute the square of the vector magnitudes, which represent the
        // multisets of the neighbors.
        double n1mag = neighbors1.magnitude();
        double n2mag = neighbors2.magnitude();
        assert !(n1mag == 0 && n2mag == 0) : "Received connections with no edges";
        double n1sq = n1mag * n1mag;
        double n2sq = n2mag * n2mag;
        
        double numer = edgeSimSum; // Math.sqrt(edgeSimSum);
        
        // This is the Tanimoto coefficient 
        double similarity = numer / (n1sq + n2sq - edgeSimSum);

        return similarity;
    }

    private <T,E extends TypedEdge<T>> double clusterSimilarity(
                Multigraph<T,E> g, IntSet cluster, RelationSemantics<T> semantics) {
        
        
        Multigraph<T,E> copy = g.copy(cluster);

        double similaritySum = 0d;
        int comparisons = 0;

        IntIterator iter1 = cluster.iterator();
        while (iter1.hasNext()) {
            int v1 = iter1.nextInt();
            
            IntSet neighbors = copy.getNeighbors(v1);
            IntIterator it1 = neighbors.iterator();
            while (it1.hasNext()) {
                int v2 = it1.nextInt();
                IntIterator it2 = neighbors.iterator();
                while (it2.hasNext()) {
                    int v3 = it2.nextInt();
                    if (v2 == v3)
                        break;
                    double sim = getConnectionSimilarity(g, v1, v2, v3, 
                                                         semantics);
                    similaritySum += sim;
                    ++comparisons;
                }
            }
        }

        return similaritySum / comparisons;
    }

    private <T,E extends TypedEdge<T>> MultiMap<Integer,Integer> 
             singleLink(final Multigraph<T,E> g, 
                        final RelationSemantics<T> semantics) {
        

        final boolean isDirected = g instanceof DirectedGraph;
        final int numTypes = g.edgeTypes().size();
        
        // Index the edges so that we can quickly look up which cluster a relation
        // is in
        final Indexer<Relation> relationIndexer = new HashIndexer<Relation>();
        // Loop over each edge in the graph and add the vertices for that edge
        // to the initial cluster assignment
        for (E edge : g.edges()) {
            Relation rel = new Relation(edge.from(), edge.to());
            relationIndexer.index(rel);
        }

        final int numRelations = relationIndexer.size();
        verbose(LOGGER, "Found %d relations in the graph with %d edges", 
                numRelations, g.size());

        // Keep a simple int->int mapping from each edge's index to (1) the
        // cluster its in, (2) the most similar edge to that edge, (3) the
        // similarity of the most similar edge.  
        final int[] relationToCluster = new int[numRelations];
        final int[] relationToMostSim = new int[numRelations];
        final double[] relationToSimOfMostSim = new double[numRelations];

        // Keep track of the vertices in each cluster
        // IntIntMultiMap clusterToVertices = new IntIntHashMultiMap();
        MultiMap<Integer,Integer> clusterToVertices
            = new HashMultiMap<Integer,Integer>();
        // IntIntMultiMap densestSolution = new IntIntHashMultiMap();
        MultiMap<Integer,Integer> densestSolution
            = new HashMultiMap<Integer,Integer>();

        //////// HACK
        MultiMap<Integer,Relation> densestSolutionRelations
            = new HashMultiMap<Integer,Relation>();        
        //////// END HACK

        // Loop over each edge in the graph and add the vertices for that edge
        // to the initial cluster assignment
        for (int i = 0; i < numRelations; ++i) {
            Relation rel = relationIndexer.lookup(i);
            int initialCluster = i;
            relationToCluster[initialCluster] = initialCluster;
            clusterToVertices.put(initialCluster, rel.v2);
            clusterToVertices.put(initialCluster, rel.v1);
        }

        // For each edge, find the most similar cluster updating the relative
        // indices of the rowToMostSimilar arrays with the results.
        Object taskKey = WORK_QUEUE.registerTaskGroup(g.order());
        IntIterator iter1 = g.vertices().iterator();
        while (iter1.hasNext()) {
            final int v1 = iter1.nextInt();
            
            WORK_QUEUE.add(taskKey, new Runnable() { 
                    public void run() {

                        veryVerbose(LOGGER, "Computing similarities for " + 
                                    "vertex %d", v1);
                        IntSet neighbors = g.getNeighbors(v1);
                        IntIterator it1 = neighbors.iterator();
                        while (it1.hasNext()) {
                            int v2 = it1.nextInt();
                            IntIterator it2 = neighbors.iterator();
                            while (it2.hasNext()) {
                                int v3 = it2.nextInt();
                                if (v2 == v3)
                                    break;
                                double sim = getConnectionSimilarity(g, v1, v2, v3, 
                                                                     semantics);
                                int r1index = relationIndexer.find(new Relation(v1, v2));
                                assert r1index >= 0 : "intruduced new unseen relation: " 
                                    + new Relation(v1, v2);
                                int r2index = relationIndexer.find(new Relation(v1, v3));
                                assert r2index >= 0 : "intruduced new unseen relation: " 
                                    + new Relation(v1, v3);
                                
                                synchronized(relationIndexer.lookup(r1index)) {
                                    if (sim > relationToSimOfMostSim[r1index]) {
                                        relationToSimOfMostSim[r1index] = sim;
                                        relationToMostSim[r1index] = r2index;
                                    }
                                }
                                synchronized(relationIndexer.lookup(r2index)) {
                                    if (sim > relationToSimOfMostSim[r2index]) {
                                        relationToSimOfMostSim[r2index] = sim;
                                        relationToMostSim[r2index] = r1index;
                                    }
                                }
                            }
                        }
                    }
                });
        }
        WORK_QUEUE.await(taskKey);

        // Keep track of the size of each cluster so that we can merge the
        // smaller into the larger.  Each cluster has an initial size of 1
        int[] clusterToNumEdges = new int[numRelations];
        Arrays.fill(clusterToNumEdges, 1);

        // As we cluster the edges, keep track of the density so that once
        // finished, we can recompute the final clustering solution.
        double highestDensity = 0d;
        int mergeStepsWithHighestDensity = 0;

        verbose(LOGGER, "Clustering edges");

        // Perform rows-1 merges to merge all elements
        int mergeIter = 0;
        while (clusterToVertices.size() > 1) {
            if (LOGGER.isLoggable(Level.FINE))
                verbose(LOGGER, "Computing dendrogram merge %d/%d",
                        mergeIter+1, numRelations-1);

            if (clusterToVertices.size() == 2) {
                // System.out.println(Arrays.toString(relationToCluster));
            }

            // Find the relation that has the highest similarity to another
            // relation
            int rel1index = -1;
            int rel2index = -1; // set during iteration
            double highestSim = -Double.MAX_VALUE;
            for (int i = 0; i < relationToSimOfMostSim.length; ++i) {
                if (relationToSimOfMostSim[i] > highestSim) {
                    int c1 = relationToCluster[i];
                    int mostSim = relationToMostSim[i];
                    int c2 = relationToCluster[mostSim];
                    if (c1 != c2) {
                        highestSim = relationToSimOfMostSim[i];
                        rel1index = i;
                        rel2index = relationToMostSim[i];
                    }
                }
            }

            int cluster1index = -1;
            int cluster2index = -1;

            // No more similar pairs (disconnected graph?) so merge two
            // arbitrary clusters and continue
            if (rel1index == -1) {
                Iterator<Integer> it = clusterToVertices.keySet().iterator();
                cluster1index = it.next();
                cluster2index = it.next(); // by contract, we have > 2 clusters
                
            }
            else {
                cluster1index = relationToCluster[rel1index];
                cluster2index = relationToCluster[rel2index];
            }

            // When merging clusters each of size > 2, it could be that
            // additional pairs of the two clusters also had higher similarity
            // and were next in the next-best-merge.  However, because these two
            // edges are already in the cluster, we can discard their merge (by
            // setting the similarity to the minimum value) and continue on.
            // Note that we do not need to update the next-most similar to these
            // edges, as that similarity will alread be recorded by the other
            // edges that are not these two.
            if (cluster1index == cluster2index) {
                relationToSimOfMostSim[rel1index] = -2d;
                // Note that since we are skipping this round, we do not
                // increment the mergeIter counter
                continue;
            }
            // We will merge this iteration
            ++mergeIter; 

            Set<Integer> verticesInSmaller = clusterToVertices.get(cluster2index);
            clusterToVertices.putMany(cluster1index, verticesInSmaller);
            clusterToVertices.remove(cluster2index);
            clusterToNumEdges[cluster1index] +=clusterToNumEdges[cluster2index];

            // System.out.println("num clusters: " + clusterToVertices.size());

            // Short circuit on the last iteration since we don't need to scan
            // through the list of edges again to update their most-similar-edge
            if (clusterToVertices.size() == 1)
                break;

            // Update the similarity for the second edge so that it is no longer
            // merged with another edge.  Even if it is more similar, we maintain
            // the invariant that only the cluster1index is valid after a merge
            // operation
            relationToSimOfMostSim[rel1index] = -4d;
            relationToSimOfMostSim[rel2index] = -5d;
            
            // For all the edges not in the current cluster, find the most
            // similar data point to the now-merged cluster.  Note that this
            // process doesn't need to update the nearest neighbors of these
            // nodes, as they should still be valid post-merge.
            int mostSimEdgeToCurCluster = -1;
            highestSim = -3d;
            Relation r1 = relationIndexer.lookup(rel1index);
            Relation r2 = relationIndexer.lookup(rel2index);

            for (int i = 0; i < numRelations; i++) {                
                int cId = relationToCluster[i];
                if (cId == cluster1index) {
                    // See if the most similar edge is also an edge in cluster1,
                    // in which case we should invalidate its similarity since
                    // the clusters are already merged
//                      int mostSimEdge = relationToMostSim[i];
//                      if (relationToCluster[mostSimEdge] == cluster2index)
//                          relationToSimOfMostSim[i] = -Double.MAX_VALUE;
                    continue;
                }
                else if (cId == cluster2index) {
                    relationToCluster[i] = cluster1index;                   
                    // See if the most similar edge is also an edge in cluster1,
                    // in which case we should invalidate its similarity since
                    // the clusters are already merged
//                      int mostSimEdge = relationToMostSim[i];
//                      if (relationToCluster[mostSimEdge] == cluster1index)
//                          relationToSimOfMostSim[i] = -Double.MAX_VALUE;
                    continue;
                }
                                
                Relation r3 = relationIndexer.lookup(i);
                double simToRel1 = getConnectionSimilarity(g, r1, r3, semantics);
                double simToRel2 = getConnectionSimilarity(g, r2, r3, semantics);
                double sim = Math.max(simToRel1, simToRel2);
                if (sim > highestSim) {
                    highestSim = sim;
                    mostSimEdgeToCurCluster = i;
                }
            }
            relationToMostSim[rel1index] = mostSimEdgeToCurCluster;
            relationToSimOfMostSim[rel1index] = highestSim;

            // Sum the densitites for each partition in this solution.
            double partitionDensitySum = 0d;
            double similaritySum = 0d;
            int totalComparisons = 0;
            for (Map.Entry<Integer,Set<Integer>> cluster : 
                     clusterToVertices.asMap().entrySet()) {   

                // Decide which edges are in this cluster
                int cid = cluster.getKey();
                Set<E> edges = new HashSet<E>();
                for (int i = 0; i < relationToCluster.length; ++i) { 
                    if (relationToCluster[i] == cid) {
                        Relation rel = relationIndexer.lookup(i);
                        edges.addAll(g.getEdges(rel.v1, rel.v2));
                    }
                }
                double simSum = 0d;
                for (E e1_ : edges) {
                    for (E e2_ : edges) {
                        if (e1_.equals(e2_)) 
                            break;
                        simSum += semantics.similarity(e1_, e2_); 
//                         System.out.printf("Similarity of %s and %s: %f%n",
//                                           e1_.edgeType(), 
//                                           e2_.edgeType(),
//                                           semantics.similarity(e1_.edgeType(), 
//                                                                e2_.edgeType()));
                    }
                }
                int size = edges.size();
                int comparisons = (size * size + size) / 2;
                
                // similaritySum += simSum / comparisons;
                similaritySum += simSum;
                totalComparisons += comparisons;

                int numNodesInCluster = cluster.getValue().size();
                    // cVertices.size();
                int numEdgesInCluster = clusterToNumEdges[cluster.getKey()];
                // Compute this partition's density, adding it to the sum of all
                // partition densities for the current solution
                if (isDirected) {
                    partitionDensitySum += computeDirectedDensity(
                        numNodesInCluster, numEdgesInCluster, numTypes);
                }
                else {
                    partitionDensitySum += computeDensity(
                        numNodesInCluster, numEdgesInCluster, numTypes);
                }
            }
            
            // Compute the total partition density by averaging the density
            // across all partitions, weighted by the number of edges
            double avgClusterSimilarity = 
                // similaritySum / clusterToVertices.size();
                similaritySum / totalComparisons;
            // Normalize the partition density by the number of edges.  In a
            // directed network, there are (V^2 - V) edges, while in an
            // undirected, there's onlt ((V^2 - V) / 2), so half the numerator
            double normalizingFactor = (isDirected) ? 1d : 2d;
            double partitionDensity = avgClusterSimilarity
                * (normalizingFactor / g.size()) * partitionDensitySum;

//             System.out.printf("Merge %d/%d had density %f%n", 
//                               mergeIter, numEdges-1, partitionDensity);

//              System.out.printf("%d %f %f %f%n", 
//                                mergeIter, partitionDensity,
//                                avgClusterSimilarity,
//                                (2d / numEdges) * partitionDensitySum);


            veryVerbose(LOGGER, "Merge %d/%d had density %f, avg. similarity: %f, " +
                        "unweight density: %f", 
                        mergeIter, numRelations-1, partitionDensity,
                        avgClusterSimilarity, (2d / g.size()) * partitionDensitySum);
            if (mergeIter % 1000 == 0)
                verbose(LOGGER, "Merge %d/%d had density %f", 
                        mergeIter, g.size()-1, partitionDensity);

            System.out.printf("FOO %d %f%n", 
                    mergeIter, partitionDensity);

            
            if (partitionDensity > highestDensity) {
                highestDensity = partitionDensity;
                mergeStepsWithHighestDensity = mergeIter;
                densestSolution.clear();
                densestSolution.putAll(clusterToVertices);
                
                densestSolutionRelations.clear();
                for (int r = 0; r < relationToCluster.length; ++r) {
                    densestSolutionRelations.put(relationToCluster[r], 
                                                 relationIndexer.lookup(r));
                }
//                 System.out.printf("Solution %d has %d clusters%n", 
//                                   mergeStepsWithHighestDensity, densestSolution.size());
            }
        }


//         System.out.printf("Merge %d had the highest density: %f%n", 
//                           mergeStepsWithHighestDensity, highestDensity);

        verbose(LOGGER, "Merge %d had the highest density: %f", 
                mergeStepsWithHighestDensity, highestDensity);

        
        for (Map.Entry<Integer,Set<Relation>> e : 
                 densestSolutionRelations.asMap().entrySet()) {
            System.out.printf("Cluster %d had relations %s%n", e.getKey(), e.getValue());
        }

        return densestSolution;
    }


    /**
     * Computes the density of the provided partition of edges
     */
    protected double computeDensity(int numNodes, int numEdges, int numTypes) {        
        // Special case when the number of nodes is 2, which has a density of 0    
        if (numNodes == 2)
            return 0;

        
         return numEdges * (numEdges - numNodes + 1d) 
             / (numTypes * (numNodes - 1d) * (numNodes - 2d));
    }

    /**
     * Computes the density of the provided partition of edges in a directed
     * network
     */
    protected double computeDirectedDensity(int numNodes, int numEdges, int numTypes) {        
        // Special case when the number of nodes is 2, which has a density of 0    
        if (numNodes == 2)
            return 0;

        int n_minus_1 = numNodes - 1;
         return numEdges * (numEdges - n_minus_1) 
             / (numTypes * n_minus_1 * n_minus_1);
    }

    /**
     * Calculates the similarity between all pair-wise combinations of edges,
     * returning a {@link PriorityQueue} sorted to have the most similar edges
     * appear at the front of the queue.  This method assumes that the only
     * similarities that matter are those that occur between two edges that
     * share a vertex.
     *
     * @param edgeIndices the list of all edges known to the system
     * @param sm a square matrix whose values denote edges between the rows.
     *
     * @return the similarity matrix
     */
    private <T,E extends TypedEdge<T>> PriorityQueue<EdgePair> 
                        calcuateEdgeSimQueue(final Multigraph<T,E> graph, 
                                             final double minSimilarity,
                                             final RelationSemantics<T> semantics) {
        
        final int numVertices = graph.order();
        final int numEdges = graph.size();
        double avgDegree = numEdges  / (double)numVertices;
        final int numComparisons = 
            (int)(((avgDegree * (avgDegree+1)) / 2) * numVertices);
        final PriorityQueue<EdgePair> pq = 
            new PriorityQueue<EdgePair>(numComparisons);
        Object key = WORK_QUEUE.registerTaskGroup(graph.order());
        
        IntIterator iter1 = graph.vertices().iterator();
        while (iter1.hasNext()) {
            final int v1 = iter1.nextInt();
            WORK_QUEUE.add(key, new Runnable() {
                    public void run() {
                        veryVerbose(LOGGER, "Computing similarities for " + 
                                    "vertex %d", v1);
                        IntSet neighbors = graph.getNeighbors(v1);
                        // Create a thread-local PriorityQueue that will hold
                        // the edge similarities for this vertex.  Once all the
                        // simialrites have been computed, we can update the
                        // thread-shared queue with minimal locking
                        PriorityQueue<EdgePair> localQ = 
                            new PriorityQueue<EdgePair>(neighbors.size());
                        IntIterator it1 = neighbors.iterator();
                        while (it1.hasNext()) {
                            int v2 = it1.nextInt();
                            
                            IntIterator it2 = neighbors.iterator();
                            while (it2.hasNext()) {
                                int v3 = it2.nextInt();
                                if (v2 == v3)
                                    break;
                                float sim = (float)
                                    getConnectionSimilarity(graph, v1, v2, v3, 
                                                            semantics);
                                if (sim > minSimilarity)
                                    localQ.add(new EdgePair(-sim, v1, v2, v3));
                            }
                        }
                        synchronized(pq) {
                            pq.addAll(localQ);
                            int comps = pq.size();
                            veryVerbose(LOGGER, "%d/%d comparisons " +
                                        "completed (%f)", comps, numComparisons,
                                        (double)comps / numComparisons);
                        }
                    }
                });
        }
        WORK_QUEUE.await(key);
        return pq;
    }

    private <T,E extends TypedEdge<T>> double getConnectionSimilarity(
                         Multigraph<T,E> graph, Relation r1, Relation r2,
                         RelationSemantics<T> semantics) {

        if (r1.v1 == r2.v1)
            return getConnectionSimilarity(
                graph, r1.v1, r1.v2, r2.v2, semantics);
        else if (r1.v1 == r2.v2)
            return getConnectionSimilarity(
                graph, r1.v1, r1.v2, r2.v1, semantics);
        else if (r1.v2 == r2.v1)
            return getConnectionSimilarity(
                graph, r1.v2, r1.v1, r2.v2, semantics);
        else if (r1.v2 == r2.v2)
            return getConnectionSimilarity(
                graph, r1.v2, r1.v1, r2.v1, semantics);
        else
            return 0;
    }

    /**
     * A structure for holding the indices of two {@link Edge} instances and
     * their corresponding negative similarity.  The similarity is negated so
     * that when sorted, the edge pair with the highest similarity appears at
     * the front of the {@link PriorityQueue}.
     */
    private static class EdgePair implements Comparable<EdgePair> {
        
        int keystone;
        int impost1;
        int impost2;

        float negSim;
        
        public EdgePair(float negSim, Edge e1, Edge e2) {
            throw new Error();
        }

        public EdgePair(float negSim, int keystone, int impost1, int impost2) {
            this.keystone = keystone;
            this.impost1 = impost1;
            this.impost2 = impost2;
            this.negSim = negSim;
        }

        public int compareTo(EdgePair ep) {
            return Float.compare(negSim, ep.negSim);
        }

        public Edge edge1() { return new SimpleEdge(keystone, impost1); }
        public Edge edge2() { return new SimpleEdge(keystone, impost2); }

        public boolean equals(Object o) {
            if (o instanceof EdgePair) {
                EdgePair ep = (EdgePair)o;
                 return keystone == ep.keystone
                     && impost1 == ep.impost1
                     && impost2 == ep.impost2
                     && negSim == ep.negSim;

            }
            return false;
        }

        public int hashCode() {
            return impost1 ^ impost2 ^ keystone;
        }

        public String toString() {
            int i = Math.min(impost1, impost2);
            int j = Math.max(impost1, impost2);
            return i + "-" + keystone + "-" + j + ": " + (-negSim);
        }
    }

    static class Relation {
        
        final int v1;
        final int v2;

        public Relation(int v1, int v2) {
            this.v1 = v1;
            this.v2 = v2;
        }
        public int hashCode() {
            return v1 ^ v2;
        }
        public boolean equals(Object o) {
            if (o instanceof Relation) {
                Relation r = (Relation)o;
                return (r.v1 == v1 && r.v2 == v2)
                    || (r.v2 == v1 && r.v1 == v2);
            }
            return false;
        }

        public String toString() {
            return "(" + v1 + "--" + v2 + ")";
        }
    }
}