package edu.ucla.sspace.graph;

import edu.ucla.sspace.common.Similarity;

import edu.ucla.sspace.clustering.Assignment;
import edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering;
import edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering.ClusterLinkage;
import edu.ucla.sspace.clustering.Merge;
import edu.ucla.sspace.clustering.SoftAssignment;

import edu.ucla.sspace.matrix.AbstractMatrix;
import edu.ucla.sspace.matrix.Matrix;
import edu.ucla.sspace.matrix.SparseHashMatrix;
import edu.ucla.sspace.matrix.SparseMatrix;
import edu.ucla.sspace.matrix.SparseSymmetricMatrix;
import edu.ucla.sspace.matrix.YaleSparseMatrix;

import edu.ucla.sspace.util.Counter;
import edu.ucla.sspace.util.HashMultiMap;
import edu.ucla.sspace.util.HashIndexer;
import edu.ucla.sspace.util.Indexer;
import edu.ucla.sspace.util.MultiMap;
import edu.ucla.sspace.util.ObjectCounter;
import edu.ucla.sspace.util.Pair;
import edu.ucla.sspace.util.WorkQueue;

import edu.ucla.sspace.util.primitive.IntIterator;
import edu.ucla.sspace.util.primitive.IntSet;
import edu.ucla.sspace.util.primitive.IntIntMultiMap;
import edu.ucla.sspace.util.primitive.IntIntHashMultiMap;
import edu.ucla.sspace.util.primitive.PrimitiveCollections;

import edu.ucla.sspace.vector.CompactSparseIntegerVector;
import edu.ucla.sspace.vector.DoubleVector;
import edu.ucla.sspace.vector.SparseDoubleVector;
import edu.ucla.sspace.vector.SparseHashIntegerVector;
import edu.ucla.sspace.vector.SparseIntegerVector;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Properties;
import java.util.Queue;
import java.util.Random;
import java.util.Set;

import java.util.concurrent.ConcurrentNavigableMap;
import java.util.concurrent.ConcurrentSkipListMap;

import java.util.logging.Level;
import java.util.logging.Logger;

import gnu.trove.iterator.TIntDoubleIterator;
import gnu.trove.map.TIntDoubleMap;
import gnu.trove.map.hash.TIntDoubleHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import static edu.ucla.sspace.util.LoggerUtil.verbose;
import static edu.ucla.sspace.util.LoggerUtil.veryVerbose;


/**
 * @author David Jurgens 
 */
public class RelationClustering3 implements java.io.Serializable {

    public static edu.ucla.sspace.util.Indexer<?> indexer;
    
    private static final long serialVersionUID = 1L;

    /**
     * A prefix for specifying properties.
     */
    private static final String PROPERTY_PREFIX =
        "edu.ucla.sspace.graph.RelationClustering3";


    private static final int DEFAULT_MAX_ITERATIONS = 1000;

    private static final double DEFAULT_RANDOM_ASSIGNMENT_PROB = 0.05d;
    
    /**
     * The logger to which clustering status updates will be written.
     */
    private static final Logger LOGGER =
        Logger.getLogger(RelationClustering3.class.getName());

    /**
     * Class local source of randomness for reproducibility.
     */
    private static final Random RANDOM = new Random();

    /**
     * Instantiates a new {@code RelationClustering3} instance.
     */
    public RelationClustering3() { }

    /**
     * <i>Ignores the specified number of clusters</i> and returns the
     * clustering solution according to the partition density.
     *
     * @param numClusters this parameter is ignored.
     */
    public <T,E extends TypedEdge<T>> MultiMap<Integer,Integer> cluster(
                        Multigraph<T,E> graph, RelationSemantics<T> semantics, 
                        Properties props) {

        // Probably should set these properties somewhere

        return cluster(graph, semantics, DEFAULT_MAX_ITERATIONS, 
                       DEFAULT_RANDOM_ASSIGNMENT_PROB);
    }

    /**
     * <i>Ignores the specified number of clusters</i> and returns the
     * clustering solution according to the partition density.
     *
     * @param numClusters this parameter is ignored.
     */
    public <T,E extends TypedEdge<T>> MultiMap<Integer,Integer> cluster(
                        Multigraph<T,E> graph, RelationSemantics<T> semantics, 
                        int maxIterations, double randomAssignmentProb) {

        return linkLabel(graph, semantics, maxIterations, randomAssignmentProb);
    }

    /**
     * Computes the similarity of the two edges as the Tanimoto coefficient of
     * the multisets of the edges connecting the neighbors of two impost nodes.
     * The impost nodes are the two nodes the edges do not have in common.
     */
    protected <T,E extends TypedEdge<T>> double getConnectionSimilarity(
            Multigraph<T,E> graph, int keystone, int impost1, int impost2,
            RelationSemantics<T> simFunc) {
        
        int verts = graph.order() + 1;

        int impost1magSum = 1; // actually 1 * 1
        int impost2magSum = 1; // actually 1 * 1

        double edgeSimSum = 0d;

        // Find all the neighbors of the first impost node
        IntIterator impost1neighborIter = 
            graph.getNeighbors(impost1).iterator();
        while (impost1neighborIter.hasNext()) {
            int neighbor = impost1neighborIter.nextInt();

            Set<E> impost1edgesToNeighbor = graph.getEdges(impost1, neighbor);
            // neighbors1.add(neighbor, impost1edgesToNeighbor.size());
            int i1nSize = impost1edgesToNeighbor.size();
            impost1magSum = i1nSize * i1nSize;

            // Examine all the other edges that the other impost node may share
            // as neighbors.  Note that this iteration will be a no-op if the
            // other impost does not have any edges to this neighbor.
            Set<E> impost2edgesToNeighbor = graph.getEdges(impost2, neighbor);
            if (impost2edgesToNeighbor.size() > 0) {
                double sim = simFunc.similarity(impost1edgesToNeighbor, 
                                                impost2edgesToNeighbor);
                edgeSimSum += sim;
                // Double count the similarity for the reverse
                if (neighbor == impost2)
                    edgeSimSum += sim;
            }
        }

        // Find all the neighbors for the second impost
        IntIterator impost2neighborIter = 
            graph.getNeighbors(impost2).iterator();
        while (impost2neighborIter.hasNext()) {
            int neighbor = impost2neighborIter.nextInt();
            Set<E> impost2edgesToNeighbor = graph.getEdges(impost2, neighbor);
            // neighbors2.add(neighbor, impost2edgesToNeighbor.size());
            int i2nSize = impost2edgesToNeighbor.size();
            impost2magSum = i2nSize * i2nSize;
        }
        
        // Compute the square of the vector magnitudes, which represent the
        // multisets of the neighbors.
        double n1mag = Math.sqrt(impost1magSum);
        double n2mag = Math.sqrt(impost2magSum);
        assert !(n1mag == 0 && n2mag == 0) : "Received connections with no edges";
        double n1sq = n1mag * n1mag;
        double n2sq = n2mag * n2mag;
        
        double numer = edgeSimSum; // Math.sqrt(edgeSimSum);
        
        // This is the Tanimoto coefficient 
        double similarity = numer / (n1sq + n2sq - edgeSimSum);

        return similarity;
    }


    private <T,E extends TypedEdge<T>> MultiMap<Integer,Integer> 
                         linkLabel(final Multigraph<T,E> g, 
                                   final RelationSemantics<T> semantics,
                                   int maxIterations, 
                                   double randomAssignmentProb) {
        
        final int numTypes = g.edgeTypes().size();
        
        // Index the edges so that we can quickly look up which cluster a
        // relation is in
        final Indexer<Relation> relationIndexer = new HashIndexer<Relation>();
        // Loop over each edge in the graph and add the vertices for that edge
        // to the initial cluster assignment
        for (E edge : g.edges()) {
            Relation rel = new Relation(edge.from(), edge.to());
            relationIndexer.index(rel);
        }

        final int numRelations = relationIndexer.size();
        verbose(LOGGER, "Found %d relations in the graph with %d edges", 
                numRelations, g.size());
        System.out.printf("Found %d relations in the graph with %d edges%n", 
                numRelations, g.size());

        // Keep a simple int->int mapping from each edge's index to (1) the
        // cluster its in, (2) the most similar edge to that edge, (3) the
        // similarity of the most similar edge.  
        final int[] relationToIterationOrder = new int[numRelations];
        final int[] relationToCluster = new int[numRelations];

        // Loop over each edge in the graph and add the vertices for that edge
        // to the initial cluster assignment
        for (int i = 0; i < numRelations; ++i) {
            relationToCluster[i] = i;
            relationToIterationOrder[i] = i;
        }
        
        // Force building the reverse mapping early
        Relation rel = relationIndexer.lookup(0);        

        boolean assignmentsChanged = true;
        double mutationRate = randomAssignmentProb;
        for (int iter = 0; iter < maxIterations && assignmentsChanged; ++iter) {
            assignmentsChanged = false;

            veryVerbose(LOGGER, "Beginning iteration %d", iter);
            System.out.printf("Beginning iteration %d%n", iter);
            
            // Shuffle the order in which the vertices will be accessed
            PrimitiveCollections.shuffle(relationToIterationOrder);

            for (int i = 0; i < relationToIterationOrder.length; ++i) {
                int curRel = relationToIterationOrder[i];

                if (i > 0 && i % 100 == 0) {
                    System.out.printf("Finished relation %d/%d (%f done)%n",
                                      i, numRelations, ((double)i) / numRelations);
                }
                
                // Allow for random mutations
                if (RANDOM.nextDouble() < mutationRate) {
                    int randomClass = RANDOM.nextInt(numRelations);
                    int oldClass = relationToCluster[curRel];
                    if (oldClass != randomClass) {
                        relationToCluster[curRel] = randomClass;
                        assignmentsChanged = true;
                    }                    
                }
                // Otherwise use the regular update procedure
                else {
                    // Get the neighbors of the current vertex and identify
                    // which class label is the maximum from the neighbors
                    int maxClass = 
                        getMaxClass(curRel, relationIndexer, 
                                    relationToCluster, g, semantics);
                    int oldClass = relationToCluster[curRel];
                    if (oldClass != maxClass) {
                        relationToCluster[curRel] = maxClass;
                        assignmentsChanged = true;
                    }
                }                                
            }
        }

        MultiMap<Integer,Integer> clusterToVertices = 
            new HashMultiMap<Integer,Integer>();
        for (int i = 0; i < relationToCluster.length; ++i) {
            int cId = relationToCluster[i];
            Relation r = relationIndexer.lookup(i);
            clusterToVertices.put(cId, r.v1);
            clusterToVertices.put(cId, r.v2);
        }
        return clusterToVertices;
    }

    /**
     * Returns the class assignment for the given relation based on the
     * assignments of its vertex.
     */
    private <T,E extends TypedEdge<T>> int getMaxClass(
                 int curRelIndex, Indexer<Relation> relationIndices,
                 int[] relationToCluster, Multigraph<T,E> g, 
                 RelationSemantics<T> semantics) {
        Relation r = relationIndices.lookup(curRelIndex);
        int v1 = r.v1;
        int v2 = r.v2;

        TIntDoubleMap classToWeight = new TIntDoubleHashMap();

        // Iterater over all the neighbors of v1, forming a second relation that
        // is adjacent
        IntIterator nIter = g.getNeighbors(v1).iterator();
        while (nIter.hasNext()) {
            int v3 = nIter.nextInt();
            double relSim = getConnectionSimilarity(g, v1, v2, v3, semantics);
            int otherRelIndex = relationIndices.find(new Relation(v1, v3));
            int otherRelClass = relationToCluster[otherRelIndex];

            // Each class is weighted according to the relation similarity of
            // this class to the other class
            if (classToWeight.containsKey(otherRelClass)) {
                double oldWeight = classToWeight.get(otherRelClass);
                classToWeight.put(otherRelClass, oldWeight + relSim);
            }
            else {
                classToWeight.put(otherRelClass, relSim);
            }
        }

        // Iterater over all the neighbors of v2, forming a second relation that
        // is adjacent to v2.
        nIter = g.getNeighbors(v2).iterator();
        while (nIter.hasNext()) {
            int v3 = nIter.nextInt();
            // Skip the relation with v1, since we would be double counting it
            if (v3 == v1)
                continue;
            double relSim = getConnectionSimilarity(g, v2, v1, v3, semantics);
            int otherRelIndex = relationIndices.find(new Relation(v2, v3));
            int otherRelClass = relationToCluster[otherRelIndex];

            // Each class is weighted according to the relation similarity of
            // this class to the other class
            if (classToWeight.containsKey(otherRelClass)) {
                double oldWeight = classToWeight.get(otherRelClass);
                classToWeight.put(otherRelClass, oldWeight + relSim);
            }
            else {
                classToWeight.put(otherRelClass, relSim);
            }
        }

//         System.out.printf("Selecting class for relation (%d, %d) " + 
//                           "from %d options%n",
//                           v1, v2,  classToWeight.size());
                          

        // Now iterate over the class weights and pick the highest classes with
        // the top weight (if there were ties)
        TIntSet ties = new TIntHashSet();
        double max = -Integer.MAX_VALUE;
        for (TIntDoubleIterator it = classToWeight.iterator(); it.hasNext(); ) {
            it.advance();
            double weight = it.value();
            if (weight > max) {
                ties.clear();
                max = weight;
            }
            if (weight == max)
                ties.add(it.key());
        }
        
        // If there were ties, randomly pick among them
        int[] options = ties.toArray();

//         System.out.printf("Selecting among tied class for relation " + 
//                           "(%d, %d): %s%n",
//                           v1, v2,  Arrays.toString(options));

        return (options.length == 1)
            ? options[0]
            : options[RANDOM.nextInt(options.length)];
    }

    /**
     * A struct representing an abstraction of a relation with an arbitrary
     * number of edges between two vertices.
     */
    static class Relation {
        
        final int v1;
        final int v2;

        public Relation(int v1, int v2) {
            this.v1 = v1;
            this.v2 = v2;
        }

        public int hashCode() {
            return v1 ^ v2;
        }

        public boolean equals(Object o) {
            if (o instanceof Relation) {
                Relation r = (Relation)o;
                return (r.v1 == v1 && r.v2 == v2)
                    || (r.v2 == v1 && r.v1 == v2);
            }
            return false;
        }

        public String toString() {
            return "(" + v1 + "--" + v2 + ")";
        }
    }
}