package edu.ucla.sspace.graph;

import edu.ucla.sspace.common.Similarity;

import edu.ucla.sspace.matrix.AbstractMatrix;
import edu.ucla.sspace.matrix.Matrix;
import edu.ucla.sspace.matrix.SparseHashMatrix;
import edu.ucla.sspace.matrix.SparseMatrix;
import edu.ucla.sspace.matrix.SparseSymmetricMatrix;
import edu.ucla.sspace.matrix.YaleSparseMatrix;

import edu.ucla.sspace.util.BoundedSortedMultiMap;
import edu.ucla.sspace.util.Counter;
import edu.ucla.sspace.util.HashMultiMap;
import edu.ucla.sspace.util.HashIndexer;
import edu.ucla.sspace.util.Indexer;
import edu.ucla.sspace.util.MultiMap;
import edu.ucla.sspace.util.ObjectCounter;
import edu.ucla.sspace.util.Pair;
import edu.ucla.sspace.util.WorkQueue;

import edu.ucla.sspace.util.primitive.IntIterator;
import edu.ucla.sspace.util.primitive.IntSet;
import edu.ucla.sspace.util.primitive.IntIntMultiMap;
import edu.ucla.sspace.util.primitive.IntIntHashMultiMap;
import edu.ucla.sspace.util.primitive.PrimitiveCollections;

import edu.ucla.sspace.vector.CompactSparseIntegerVector;
import edu.ucla.sspace.vector.DoubleVector;
import edu.ucla.sspace.vector.SparseDoubleVector;
import edu.ucla.sspace.vector.SparseHashIntegerVector;
import edu.ucla.sspace.vector.SparseIntegerVector;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Properties;
import java.util.Queue;
import java.util.Random;
import java.util.Set;

import java.util.concurrent.ConcurrentNavigableMap;
import java.util.concurrent.ConcurrentSkipListMap;

import java.util.concurrent.atomic.AtomicLong;

import java.util.logging.Level;
import java.util.logging.Logger;

import gnu.trove.iterator.TIntDoubleIterator;
import gnu.trove.list.TIntList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.map.TIntDoubleMap;
import gnu.trove.map.hash.TIntDoubleHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import static edu.ucla.sspace.util.LoggerUtil.verbose;
import static edu.ucla.sspace.util.LoggerUtil.veryVerbose;


/**
 * @author David Jurgens 
 */
public class RelationClustering7 implements java.io.Serializable {

    public static edu.ucla.sspace.util.Indexer<?> indexer;
    
    private static final long serialVersionUID = 1L;

    /**
     * A prefix for specifying properties.
     */
    private static final String PROPERTY_PREFIX =
        "edu.ucla.sspace.graph.RelationClustering7";


    private static final int DEFAULT_MAX_ITERATIONS = 1000;

    private static final double DEFAULT_RANDOM_ASSIGNMENT_PROB = 0.05d;
    
    /**
     * The logger to which clustering status updates will be written.
     */
    private static final Logger LOGGER =
        Logger.getLogger(RelationClustering7.class.getName());

    /**
     * The work used by all {@code LinkClustering} instances to perform
     * multi-threaded operations.
     */
    private static final WorkQueue WORK_QUEUE = WorkQueue.getWorkQueue(8);

    /**
     * Class local source of randomness for reproducibility.
     */
    private static final Random RANDOM = new Random();

    private static final boolean DEBUG = false;

    private java.io.PrintWriter assignmentLog;

    private final double alpha;

    /**
     * Instantiates a new {@code RelationClustering7} instance.
     */
    public RelationClustering7() { 
        this(1);
    }

    /**
     * Instantiates a new {@code RelationClustering7} instance.
     */
    public RelationClustering7(double alpha) { 
        this.alpha = alpha;
        if (DEBUG) {
            try {
                assignmentLog = new java.io.PrintWriter(
                    "RC-overlap-weighted+multiple+assignments-" + RANDOM.nextInt(10000) + ".txt");
            } catch (java.io.IOException ioe) {
                throw new java.io.IOError(ioe);
            }
        }        
    }

    /**
     * <i>Ignores the specified number of clusters</i> and returns the
     * clustering solution according to the partition density.
     *
     * @param numClusters this parameter is ignored.
     */
    public  MultiMap<Integer,Integer> cluster(
                        UndirectedMultigraph<SparseDoubleVector> graph, RelationSemantics<SparseDoubleVector> semantics, 
                        Properties props) {

        // Probably should set these properties somewhere

        return cluster(graph, semantics, DEFAULT_MAX_ITERATIONS, 
                       DEFAULT_RANDOM_ASSIGNMENT_PROB);
    }

    /**
     * <i>Ignores the specified number of clusters</i> and returns the
     * clustering solution according to the partition density.
     *
     * @param numClusters this parameter is ignored.
     */
    public  MultiMap<Integer,Integer> cluster(
                        UndirectedMultigraph<SparseDoubleVector> graph, RelationSemantics<SparseDoubleVector> semantics, 
                        int maxIterations, double randomAssignmentProb) {

        return linkLabel(graph, semantics, maxIterations, randomAssignmentProb);
    }

    private  double getConnectionSimilarity(
                         UndirectedMultigraph<SparseDoubleVector> graph, Relation r1, Relation r2,
                         RelationSemantics<SparseDoubleVector> semantics) {

        if (r1.v1 == r2.v1)
            return getConnectionSimilarity(
                graph, r1.v1, r1.v2, r2.v2, semantics);
        else if (r1.v1 == r2.v2)
            return getConnectionSimilarity(
                graph, r1.v1, r1.v2, r2.v1, semantics);
        else if (r1.v2 == r2.v1)
            return getConnectionSimilarity(
                graph, r1.v2, r1.v1, r2.v2, semantics);
        else if (r1.v2 == r2.v2)
            return getConnectionSimilarity(
                graph, r1.v2, r1.v1, r2.v1, semantics);
        else
            return 0;
    }

    /**
     * Computes the similarity of the two edges as the Tanimoto coefficient of
     * the multisets of the edges connecting the neighbors of two impost nodes.
     * The impost nodes are the two nodes the edges do not have in common.
     */
    protected  double getConnectionSimilarity(
            UndirectedMultigraph<SparseDoubleVector> graph, int keystone, int impost1, int impost2,
            RelationSemantics<SparseDoubleVector> simFunc) {
        
//         System.out.printf("Computing similarity of (%d, %d) and (%d, %d)%n",
//                           keystone, impost1, keystone, impost2);

        int impost1magSum = 1; // actually 1 * 1
        int impost2magSum = 1; // actually 1 * 1

        double edgeSimSum = 0d;

        int numSharedEdges = 0;

        // Find all the neighbors of the first impost node
        IntIterator impost1neighborIter = 
            graph.getNeighbors(impost1).iterator();
        while (impost1neighborIter.hasNext()) {
            int neighbor = impost1neighborIter.nextInt();

            Set<TypedEdge<SparseDoubleVector>> impost1edgesToNeighbor = 
                graph.getEdges(impost1, neighbor);
            // neighbors1.add(neighbor, impost1edgesToNeighbor.size());
            int i1nSize = 0;
            for (TypedEdge<SparseDoubleVector> e : impost1edgesToNeighbor)
                i1nSize += e.edgeType().getNonZeroIndices().length;
            // int i1nSize = impost1edgesToNeighbor.size();
            impost1magSum += i1nSize * i1nSize;

            // Examine all the other edges that the other impost node may share
            // as neighbors.  Note that this iteration will be a no-op if the
            // other impost does not have any edges to this neighbor.
            Set<TypedEdge<SparseDoubleVector>> impost2edgesToNeighbor = 
                graph.getEdges(impost2, neighbor);
            if (impost2edgesToNeighbor.size() > 0) {
                double sim = simFunc.similarity(impost1edgesToNeighbor, 
                                                impost2edgesToNeighbor);

                int[] i1types = impost1edgesToNeighbor
                    .iterator().next().edgeType().getNonZeroIndices();
                int[] i2types = impost2edgesToNeighbor
                    .iterator().next().edgeType().getNonZeroIndices();
                numSharedEdges += getOverlap(i1types, i2types);

//                 System.out.printf("Similarity of edge set %s to %s: %f%n",
//                                   impost1edgesToNeighbor, 
//                                   impost2edgesToNeighbor, sim);
                edgeSimSum += sim;
                // Double count the similarity for the reverse
                if (neighbor == impost2)
                    edgeSimSum += sim;
            }
        }

        // Find all the neighbors for the second impost
        IntIterator impost2neighborIter = 
            graph.getNeighbors(impost2).iterator();
        while (impost2neighborIter.hasNext()) {
            int neighbor = impost2neighborIter.nextInt();
            Set<TypedEdge<SparseDoubleVector>> impost2edgesToNeighbor = 
                graph.getEdges(impost2, neighbor);
            // neighbors2.add(neighbor, impost2edgesToNeighbor.size());
            // int i2nSize = impost2edgesToNeighbor.size();
            int i2nSize = 0;
            for (TypedEdge<SparseDoubleVector> e : impost2edgesToNeighbor)
                i2nSize += e.edgeType().getNonZeroIndices().length;

            impost2magSum += i2nSize * i2nSize;
        }
        
        // Compute the square of the vector magnitudes, which represent the
        // multisets of the neighbors.
        double n1mag = Math.sqrt(impost1magSum);
        double n2mag = Math.sqrt(impost2magSum);
//         System.out.printf("magSum: %f, magSum^2: %d, magSum: %f, magSum^2: %d%n",
//                           n1mag, impost1magSum, n2mag, impost2magSum);

        assert !(n1mag == 0 && n2mag == 0) : "Received connections with no edges";
        double n1sq = n1mag * n1mag;
        double n2sq = n2mag * n2mag;
        
        double numer = edgeSimSum; // Math.sqrt(edgeSimSum);
        
        // This is the Tanimoto coefficient 
        double similarity = (numer + (numSharedEdges * alpha)) / (n1sq + n2sq - edgeSimSum);
//         System.out.printf("Similarity of (%d, %d) and (%d, %d): %f%n",
//                           keystone, impost1, keystone, impost2, similarity);

        return similarity;
    }

    private static int getOverlap(int[] arr1, int[] arr2) {
        arr1 = arr1.clone();
        Arrays.sort(arr1);
        int overlaps = 0;
        for (int i : arr2) {
            int idx = Arrays.binarySearch(arr1, i);
            if (idx >= 0)
                overlaps++;
        }
        return overlaps;
    }

    private  MultiMap<Integer,Integer> 
        linkLabel(final UndirectedMultigraph<SparseDoubleVector> g, 
                  final RelationSemantics<SparseDoubleVector> semantics,
                  int maxIterations, 
                  double randomAssignmentProb) {
        
        final int numTypes = g.edgeTypes().size();
        
        // Index the edges so that we can quickly look up which cluster a
        // relation is in
        final Indexer<Relation> relationIndexer = new HashIndexer<Relation>();
        // Loop over each edge in the graph and add the vertices for that edge
        // to the initial cluster assignment
        for (TypedEdge<?> edge : g.edges()) {
            Relation rel = new Relation(edge.from(), edge.to());
            relationIndexer.index(rel);
        }

        final int numRelations = relationIndexer.size();
        verbose(LOGGER, "Found %d relations in the graph with %d edges", 
                numRelations, g.size());
        System.out.printf("Found %d relations in the graph with %d edges%n", 
                numRelations, g.size());

        // Keep a simple int->int mapping from each edge's index to (1) the
        // cluster its in, (2) the most similar edge to that edge, (3) the
        // similarity of the most similar edge.  
        final int[] relationToIterationOrder = new int[numRelations];
        final int[] relationToCluster = new int[numRelations];

        // Keep a mapping from a relation index to the index of its neighboring
        // relations to which it is connected
        final int[][] relationToNeighbors = new int[numRelations][0];

        // Force building the reverse mapping early before the multithreaded code
        Relation rel = relationIndexer.lookup(0);        

        // Loop over each edge in the graph and add the vertices for that edge
        // to the initial cluster assignment
        Object taskKey = WORK_QUEUE.registerTaskGroup(numRelations);
        for (int i_ = 0; i_ < numRelations; ++i_) {
            final int i = i_;
            WORK_QUEUE.add(taskKey, new Runnable() {
                    public void run() {
                        relationToCluster[i] = i;
                        relationToIterationOrder[i] = i;
                        relationToNeighbors[i] = 
                            getNeighbors(g, i, relationIndexer);
                        if (i > 0 && i % 10000 == 0) {
                            System.out.printf("Finished neighbor update %d/%d "+
                                              "(%f done)%n", i, numRelations, 
                                              ((double)i) / numRelations);
                        }
                    }
                });
        }
        WORK_QUEUE.await(taskKey);
    
        
        // Build a relation-relation similarity matrix to avoid recomputing this
        // frequently
        final AtomicLong numRelComparisons = new AtomicLong();
        final AtomicLong nonZeroSimilarities = new AtomicLong();
        final SparseMatrix similarityMatrix = 
            new SparseSymmetricMatrix(new SparseHashMatrix(numRelations, numRelations));
        taskKey = WORK_QUEUE.registerTaskGroup(numRelations);
        for (int i_ = 0; i_ < numRelations; ++i_) {
            final int i = i_;
            WORK_QUEUE.add(taskKey, new Runnable() {
                    public void run() {
            
                        Relation r1 = relationIndexer.lookup(i);
                        if (i > 0 && i % 10000 == 0) {
                            System.out.printf("Finished matrix update %d/%d (%f done)%n",
                                              i, numRelations, ((double)i) / numRelations);
                        }
                        
                        for (int j : relationToNeighbors[i]) {
                            // Because this is a symmetric similarity, we can skip half of
                            // the comparisons
                            if (i < j) 
                                continue;
                            Relation r2 = relationIndexer.lookup(j);
                            double sim = getConnectionSimilarity(g, r1, r2, semantics);
                            if (sim > 0) {
                                nonZeroSimilarities.incrementAndGet();
                                // Lock the matrix for updating so we don't
                                // modify the backing vectors inconsistently.
                                // We expect that the overhead of the similarity
                                // computation is substatially more than the
                                // locking, so the process won't block too often
                                synchronized(similarityMatrix) {
                                    similarityMatrix.set(i, j, sim);
                                }
                            }
                            numRelComparisons.incrementAndGet();
                        }
                    }
                });
        }
        WORK_QUEUE.await(taskKey);

        System.out.printf("Saw %d non-zero relation similarities (%f). " +
                          "Matrix is %f dense%n", nonZeroSimilarities.get(),
                          nonZeroSimilarities.doubleValue() /  numRelComparisons.get(),
                          nonZeroSimilarities.doubleValue() 
                              / (numRelations * numRelations - numRelations));

        int assignmentsChanged = 1;
        double mutationRate = randomAssignmentProb;
        for (int iter = 0; iter < maxIterations && assignmentsChanged > 0; ++iter) {
            assignmentsChanged = 0;

            veryVerbose(LOGGER, "Beginning iteration %d", iter);
            System.out.printf("Beginning iteration %d%n", iter);
            
            // Shuffle the order in which the vertices will be accessed
            PrimitiveCollections.shuffle(relationToIterationOrder);

            for (int i = 0; i < relationToIterationOrder.length; ++i) {
                int curRel = relationToIterationOrder[i];

                if (i > 0 && i % 10000 == 0) {
                    System.out.printf("Finished relation %d/%d " + 
                                      "(%f done; iter: %d, change: %d)%n",
                                      i, numRelations, ((double)i) / numRelations,
                                      i, assignmentsChanged);
                }
                
                // Allow for random mutations
                if (RANDOM.nextDouble() < mutationRate) {
                    int randomClass = RANDOM.nextInt(numRelations);
                    int oldClass = relationToCluster[curRel];
                    if (oldClass != randomClass) {
                        relationToCluster[curRel] = randomClass;
                        assignmentsChanged++;
                    }                    
                }
                // Otherwise use the regular update procedure
                else {
                    // Get the neighbors of the current vertex and identify
                    // which class label is the maximum from the neighbors
                    int maxClass = 
                        getMaxClass(curRel, relationIndexer, 
                                    relationToCluster, similarityMatrix,
                                    relationToNeighbors);
                    int oldClass = relationToCluster[curRel];
                    if (oldClass != maxClass) {
                        relationToCluster[curRel] = maxClass;
                        assignmentsChanged++;
                    }
                }                                
            }

            if (DEBUG) {
                assignmentLog.println(Arrays.toString(relationToCluster));
            }

            if (1 == 1) {
                Counter<Integer> c = new ObjectCounter<Integer>();
                for (int i : relationToCluster)
                    c.count(i);
                System.out.printf("Saw %d classes%n", c.size());
                BoundedSortedMultiMap<Integer,Integer> b = 
                    new BoundedSortedMultiMap<Integer,Integer>(10);
                for (Map.Entry<Integer,Integer> e : c)
                    b.put(e.getValue(), e.getKey());
                for (Integer f : b.keySet()) {
                    System.out.printf("The following classes appeared %d times: %s%n",
                                      f, b.get(f));
                }
            }
        }

        if (DEBUG) {
//             try {
                assignmentLog.close();
//             } catch (java.io.IOException ioe) {
//                 throw new java.io.IOError(ioe);
//             }
        }

        MultiMap<Integer,Integer> clusterToVertices = 
            new HashMultiMap<Integer,Integer>();
        for (int i = 0; i < relationToCluster.length; ++i) {
            int cId = relationToCluster[i];
            Relation r = relationIndexer.lookup(i);
            clusterToVertices.put(cId, r.v1);
            clusterToVertices.put(cId, r.v2);
        }
        return clusterToVertices;
    }

    private int[] getNeighbors(
                         UndirectedMultigraph<SparseDoubleVector> g, int relIndex, 
                         Indexer<Relation> relationIndices) {

        Relation r = relationIndices.lookup(relIndex);
        int v1 = r.v1;
        int v2 = r.v2;
        TIntList neighbors = new TIntArrayList();

        // Iterater over all the neighbors of v1, forming a second relation that
        // is adjacent
        IntIterator nIter = g.getNeighbors(v1).iterator();
        while (nIter.hasNext()) {
            int v3 = nIter.nextInt();
            // Skip comparing the relation with itself
            if (v3 == v2)
                continue;
            int otherRelIndex = relationIndices.find(new Relation(v1, v3));
            assert otherRelIndex >= 0
                : "missing relation: " + new Relation(v1, v3);
            neighbors.add(otherRelIndex);
        }

        // Iterater over all the neighbors of v2, forming a second relation that
        // is adjacent to v2.
        nIter = g.getNeighbors(v2).iterator();
        while (nIter.hasNext()) {
            int v3 = nIter.nextInt();
            // Skip comparing the relation with itself
            if (v3 == v1)
                continue;
            int otherRelIndex = relationIndices.find(new Relation(v2, v3));
            assert otherRelIndex >= 0
                : "missing relation: " + new Relation(v2, v3);
            neighbors.add(otherRelIndex);
        }
        return neighbors.toArray();
    }

    /**
     * Returns the class assignment for the given relation based on the
     * assignments of its vertex.
     */
    private int getMaxClass(int curRelIndex, Indexer<Relation> relationIndices,
                            int[] relationToCluster, 
                            SparseMatrix similarityMatrix, 
                            int[][] relationToNeighbors) {

        int[] neighborIndices = relationToNeighbors[curRelIndex];
        TIntDoubleMap classToWeight = 
            new TIntDoubleHashMap(neighborIndices.length);

        for (int n : neighborIndices) {
            double relSim = similarityMatrix.get(curRelIndex, n);
            int otherRelClass = relationToCluster[n];

            // Each class is weighted according to the relation similarity of
            // this class to the other class
            if (classToWeight.containsKey(otherRelClass)) {
                double oldWeight = classToWeight.get(otherRelClass);
                classToWeight.put(otherRelClass, oldWeight + relSim);
            }
            else {
                classToWeight.put(otherRelClass, relSim);
            }
        }
                          

        // Now iterate over the class weights and pick the highest classes with
        // the top weight (if there were ties)
        TIntSet ties = new TIntHashSet(128);
        double max = -Integer.MAX_VALUE;
        for (TIntDoubleIterator it = classToWeight.iterator(); it.hasNext(); ) {
            it.advance();
            double weight = it.value();
            if (weight > max) {
                ties.clear();
                max = weight;
            }
            if (weight == max)
                ties.add(it.key());
        }
        
        // If this relation was isolated from all the other relations (not
        // similar to anything) or was disconnected from the rest of the
        // network, then return its class
        if (ties.isEmpty())
            return relationToCluster[curRelIndex];

        // If there were ties, randomly pick among them
        int[] options = ties.toArray();
        
//         System.out.printf("Selecting among tied class for relation " + 
//                           "(%d, %d): %s%n",
//                           v1, v2,  Arrays.toString(options));
        
        return (options.length == 1)
            ? options[0]
            : options[RANDOM.nextInt(options.length)];
    }

    /**
     * A struct representing an abstraction of a relation with an arbitrary
     * number of edges between two vertices.
     */
    static class Relation {
        
        final int v1;
        final int v2;

        public Relation(int v1, int v2) {
            this.v1 = v1;
            this.v2 = v2;
        }

        public int hashCode() {
            return v1 ^ v2;
        }

        public boolean equals(Object o) {
            if (o instanceof Relation) {
                Relation r = (Relation)o;
                return (r.v1 == v1 && r.v2 == v2)
                    || (r.v2 == v1 && r.v1 == v2);
            }
            return false;
        }

        public String toString() {
            return "(" + v1 + "--" + v2 + ")";
        }
    }
}