/*
 * Copyright 2012 David Jurgens
 *
 * This file is part of the S-Space package and is covered under the terms and
 * conditions therein.
 *
 * The S-Space package is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation and distributed hereunder to you.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND NO REPRESENTATIONS OR WARRANTIES,
 * EXPRESS OR IMPLIED ARE MADE.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, WE MAKE
 * NO REPRESENTATIONS OR WARRANTIES OF MERCHANT- ABILITY OR FITNESS FOR ANY
 * PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE OR DOCUMENTATION
 * WILL NOT INFRINGE ANY THIRD PARTY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER
 * RIGHTS.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package edu.ucla.sspace.graph.io;

import edu.ucla.sspace.graph.DirectedEdge;
import edu.ucla.sspace.graph.DirectedGraph;
import edu.ucla.sspace.graph.DirectedMultigraph;
import edu.ucla.sspace.graph.DirectedTypedEdge;
import edu.ucla.sspace.graph.Edge;
import edu.ucla.sspace.graph.Multigraph;
import edu.ucla.sspace.graph.Graph;
import edu.ucla.sspace.graph.SimpleDirectedEdge;
import edu.ucla.sspace.graph.SimpleDirectedTypedEdge;
import edu.ucla.sspace.graph.SimpleEdge;
import edu.ucla.sspace.graph.SimpleTypedEdge;
import edu.ucla.sspace.graph.SimpleWeightedEdge;
import edu.ucla.sspace.graph.SparseDirectedGraph;
import edu.ucla.sspace.graph.SparseWeightedGraph;
import edu.ucla.sspace.graph.SparseUndirectedGraph;
import edu.ucla.sspace.graph.TypedEdge;
import edu.ucla.sspace.graph.UndirectedMultigraph;
import edu.ucla.sspace.graph.WeightedEdge;
import edu.ucla.sspace.graph.WeightedGraph;

import edu.ucla.sspace.util.HashIndexer;
import edu.ucla.sspace.util.Indexer;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;

import java.util.EnumMap;
import java.util.Map;

import java.util.logging.Level;
import java.util.logging.Logger;

import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import static edu.ucla.sspace.util.LoggerUtil.verbose;
import static edu.ucla.sspace.util.LoggerUtil.veryVerbose;


/**
 * A class for reading graph files that are formatted as comma separated value
 * (csv) files.  Columns are interpreted according to the reader's construction
 * using the {@link CsvFileReader#CsvColumnType} enum.
 */
public class CsvFileReader extends GraphReaderAdapter implements GraphReader {

    /**
     * The logger to which clustering status updates will be written.
     */
    private static final Logger LOGGER =
        Logger.getLogger(CsvFileReader.class.getName());

    public enum CsvColumnType { VERTEX1, VERTEX2, WEIGHT, EDGE_TYPE };

    private final Map<CsvColumnType,Integer> columnTypes;

    public CsvFileReader() { 
        this(getDefaultColumns());
    }
    
    public CsvFileReader(Map<CsvColumnType,Integer> columnTypes) { 
        this.columnTypes = columnTypes;
        // Check that columns aren't mapped to the same value
        TIntSet vals = new TIntHashSet(columnTypes.values());
        if (vals.size() != columnTypes.size())
            throw new IllegalArgumentException(
                "Cannot map column types to the same column");
        if (!columnTypes.containsKey(CsvColumnType.VERTEX1)
               || !columnTypes.containsKey(CsvColumnType.VERTEX2)) {
            throw new IllegalArgumentException(
                "Must specify at least the columns for the vertices");
        }
    }

    /**
     * Returns the default column mapping for the reader
     */
    static Map<CsvColumnType,Integer> getDefaultColumns() {
        Map<CsvColumnType,Integer> columnTypes = 
            new EnumMap<CsvColumnType,Integer>(CsvColumnType.class);
        int i = 0;
        for (CsvColumnType c : CsvColumnType.values())
            columnTypes.put(c, i++);
        return columnTypes;
    }

    @Override public Graph<Edge> readUndirected(File f) throws IOException {
        return readUndirected(f, new HashIndexer<String>());
    }

    @Override public Graph<Edge> readUndirected(File f, Indexer<String> vertexIndexer)
            throws IOException {        
        BufferedReader br = new BufferedReader(new FileReader(f));
        Graph<Edge> g = //new GenericGraph<Edge>();
            new SparseUndirectedGraph();

        int v1col = columnTypes.get(CsvColumnType.VERTEX1);
        int v2col = columnTypes.get(CsvColumnType.VERTEX2);

        int lineNo = 0;
        for (String line = null; (line = br.readLine()) != null; ) {
            ++lineNo;
            line = line.trim();
            if (line.startsWith("#"))
                continue;
            else if (line.length() == 0)
                continue;
            String[] arr = line.split(",");
            if (arr.length < 2) {
                throw new IOException("Missing vertex on line " + lineNo);
            }
            int v1 = vertexIndexer.index(arr[v1col]);
            int v2 = vertexIndexer.index(arr[v2col]);
            g.add(new SimpleEdge(v1, v2));
            if (lineNo % 100000 == 0)
                verbose(LOGGER, "Read %d lines from %s", lineNo, f);
        }
        verbose(LOGGER, "Read undirected graph with %d vertices and %d edges", 
                g.order(), g.size());
        return g;
    }
}
