/*
 * Copyright 2012 David Jurgens
 *
 * This file is part of the S-Space package and is covered under the terms and
 * conditions therein.
 *
 * The S-Space package is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation and distributed hereunder to you.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND NO REPRESENTATIONS OR WARRANTIES,
 * EXPRESS OR IMPLIED ARE MADE.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, WE MAKE
 * NO REPRESENTATIONS OR WARRANTIES OF MERCHANT- ABILITY OR FITNESS FOR ANY
 * PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE OR DOCUMENTATION
 * WILL NOT INFRINGE ANY THIRD PARTY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER
 * RIGHTS.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package edu.ucla.sspace.util;

import java.awt.Color;

import java.util.Iterator;
import java.util.Random;


/**
 * A utility for generating a random sequence of colors that have a minimum
 * amount of contrast between them.  Contrast is determined according to <a
 * href="http://juicystudio.com/article/luminositycontrastratioalgorithm.php">this
 * algorithm</a>.  According to the <a
 * href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast">Web Content
 * Accessibility Guidelines 2.0</a>, the recommended minimum contrast ratio is 5
 * at a minimum and at least 10 for high conformance.
 *
 * <p> This class is designed for cases where an arbitrary number of contrasting
 * series of colors is needed, e.g., a stacked chart where the lines must be
 * distinguishable from each other, or where colors are needed to distinguish
 * foreground text from a  background.
 */
public class ContrastColorGenerator implements Iterator<Color> {

    private final Random rand;

    private Color prev;

    private final double minContrastRatio;

    public ContrastColorGenerator(double minContrastRatio, Color seed) { 
        if (minContrastRatio < 1 || minContrastRatio > 21)
            throw new IllegalArgumentException(
                "Contrast rations must be within the range [1:21]");
        if (seed == null)
            throw new NullPointerException("seed cannot be null");
        this.minContrastRatio = minContrastRatio;
        rand = new Random();
        prev = seed;
    }

    public ContrastColorGenerator(double minContrastRatio) { 
        this(minContrastRatio, 
             new Color((float) Math.random(),
                       (float) Math.random(),
                       (float) Math.random()));
    }

    /**
     * Always returns {@code true}.
     */
    public boolean hasNext() {
        return true;
    }

    public Color next() {
        Color next = null;

        double prevLuminosity = 
            0.2126* (prev.getRed() / 255d) 
            + .7152 * (prev.getGreen() / 255d)
            + .0722 * (prev.getBlue() / 255d);

        while (next == null) {

            // Generate a candidate color
            float r = rand.nextFloat();
            float g = rand.nextFloat();
            float b = rand.nextFloat();

            // Calculate the brightness of the new color
            double curLuminosity = 0.2126 * r + .7152 * g + .0722 * b;

            // Calculate the luminosity contrast ratio
            double min = Math.min(curLuminosity, prevLuminosity);
            double max = Math.max(curLuminosity, prevLuminosity);
            
            double ratio = max / min;
            if (ratio >= minContrastRatio) { 
                next = new Color(r, g, b);
                prev = next;
            }
        }
        return next;
    }
  
    /**
     * Returns the next random color
     */
    private Color nextColor() {
        int r = rand.nextInt(256);
        int g = rand.nextInt(256);
        int b = rand.nextInt(256);
        return new Color(r, g, b);
    }
    
    /**
     * Throws an {@link UnsupportedOperationException} if called.
     */
    public void remove() {
        throw new UnsupportedOperationException("Cannot remove colors");
    }
}